
#define ExitMenu_BGColor (RGB15(8,8,8)|BIT15)

#define ExitMenu_ItemsCount (5)
static const char *ExitMenu_pItems[ExitMenu_ItemsCount]={
  "Cancel.",
  "Overwrite save and exit.",
  "Save as name and exit.",
  "Create new filename and exit.",
  "Exit without saving.",
};
static u32 ExitMenu_ItemIndex;

static void ExitMenu_DrawScreen_DrawText(CglCanvas *pcan,u32 x,u32 y,u16 col,const char *pstr)
{
  pcan->SetFontTextColor(RGB15(0,0,0)|BIT15);
  pcan->TextOutA(x+1,y+1,pstr);
  pcan->SetFontTextColor(col);
  pcan->TextOutA(x+0,y+0,pstr);
}

static void ExitMenu_DrawScreen(CglCanvas *pcan)
{
  pcan->FillFull(ExitMenu_BGColor);
  
  const u16 TextColor=RGB15(24,24,24)|BIT15;
  const u16 TextHighColor=RGB15(31,31,31)|BIT15;
  
  u32 tx=16;
  u32 ty=16;
  const u32 th=glCanvasTextHeight+2;
  
  ty+=th;
  
  ExitMenu_DrawScreen_DrawText(pcan,tx,ty,TextColor,"EXIT MENU");
  ty+=th*2;
  
  for(u32 idx=0;idx<ExitMenu_ItemsCount;idx++){
    const char *pstr=ExitMenu_pItems[idx];
    if(idx==ExitMenu_ItemIndex){
      ExitMenu_DrawScreen_DrawText(pcan,tx+0,ty,TextHighColor,">");
      ExitMenu_DrawScreen_DrawText(pcan,tx+8,ty,TextHighColor,pstr);
      }else{
      ExitMenu_DrawScreen_DrawText(pcan,tx+8,ty,TextColor,pstr);
    }
    ty+=th;
  }
  
}

static bool ExitMenu_KeyPress(u16 keys)
{
  if((keys&(KEY_UP|KEY_DOWN|KEY_LEFT|KEY_RIGHT))!=0){
    PlaySE_Short();
    if((keys&KEY_UP)!=0){
      if(0<ExitMenu_ItemIndex) ExitMenu_ItemIndex--;
    }
    if((keys&KEY_DOWN)!=0){
      if(ExitMenu_ItemIndex<(ExitMenu_ItemsCount-1)) ExitMenu_ItemIndex++;
    }
    if((keys&KEY_LEFT)!=0){
      ExitMenu_ItemIndex=0;
    }
    if((keys&KEY_RIGHT)!=0){
      ExitMenu_ItemIndex=ExitMenu_ItemsCount-1;
    }
    return(false);
  }
  
  if((keys&(KEY_START|KEY_SELECT|KEY_A))!=0){
    return(true);
  }
  
  return(false);
}

static void ExitMenu_SaveToFile(EFilenameType FilenameType)
{
  Sound_Start(WAVFN_Opening);
  
  CglCanvas *pcan=pScreenSub->pCanvas;
  pcan->SetColor(RGB15(0,0,0)|BIT15);
  pcan->FillBox(0,0,ScreenWidth,16*4);
  pcan->SetFontTextColor(RGB15(31,31,31)|BIT15);
  pcan->TextOutA(2,(16*1)+2,"Save text file and reboot...");
  pcan->TextOutA(2,(16*2)+2,"Please don't turn off power.");
  
  TextEdit_SaveToFile(FilenameType);
}

static bool ExitMenu_Execute(void)
{
  pScreenMain->pBackCanvas->FillFull(ExitMenu_BGColor);
  pScreenMain->Flip(true);
  
  ExitMenu_ItemIndex=0;
  
  ExitMenu_DrawScreen(pScreenSub->pCanvas);
  
  while(1){
    u16 keys=(~REG_KEYINPUT)&0x3ff;
    if(keys==0) break;
    swiWaitForVBlank();
  }
  
  while(1){
    u16 keys;
    while(1){
      keys=(~REG_KEYINPUT)&0x3ff;
      if(keys!=0) break;
      swiWaitForVBlank();
    }
    bool execute=false;
    bool cancel=false;
    if((keys&KEY_B)!=0){
      cancel=true;
      }else{
      if(ExitMenu_KeyPress(keys)==true) execute=true;
    }
    ExitMenu_DrawScreen(pScreenSub->pCanvas);
    while(1){
      keys=(~REG_KEYINPUT)&0x3ff;
      if(keys==0) break;
      swiWaitForVBlank();
    }
    if(cancel==true){
      PlaySE_Long();
      return(false);
    }
    if(execute==true){
      switch(ExitMenu_ItemIndex){
        case 0: PlaySE_Long(); return(false); break;
        case 1: ExitMenu_SaveToFile(EFT_Overwrite); return(true); break;
        case 2: ExitMenu_SaveToFile(EFT_SaveAs); return(true); break;
        case 3: ExitMenu_SaveToFile(EFT_NewFilename); return(true); break;
        case 4: PlaySE_Long(); return(true); break;
      }
      return(false);
    }
  }
  
  return(false);
}
